import numpy as np
from pymoo.core.population import Population

from pydmoo.algorithms.base.dmoo.dnsga2 import DNSGA2
from pydmoo.core.ar_model import ARModel
from pydmoo.core.bounds import matrix_conditional_update
from pydmoo.core.manifold import manifold_prediction


class NSGA2PPS(DNSGA2):
    """Population Prediction Strategy (Center point prediction and manifold prediction).

    References
    ----------
    Zhou, A., Jin, Y., and Zhang, Q. (2014).
    A population prediction strategy for evolutionary dynamic multiobjective optimization.
    IEEE Transactions on Cybernetics, 44(1), 40–53.
    https://doi.org/10.1109/TCYB.2013.2245892
    """

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.p = 3  # the order of the AR model
        self.M = 23  # the length of history mean point series

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # archive center points
        center_points = self.data.get("center_points", [])
        center_points.append(np.mean(self.opt.get("X"), axis=0))

        # the maximum length
        center_points = center_points[(-self.M):]
        self.data["center_points"] = center_points

        # archive populations
        Xs = self.data.get("Xs", [])
        Xs.append(self.pop.get("X"))  # pop
        Xs = Xs[-2:]
        self.data["Xs"] = Xs

        if len(center_points) >= (self.p + 1):

            C1, distance = manifold_prediction(Xs[0], Xs[1])
            n = C1.shape[1]  # Dimensionality of the manifold
            variance = (distance ** 2) / n

            center, variances = self.center_points_prediction(center_points)

            X = center + C1 + self.random_state.normal(loc=0, scale=np.sqrt(variances + variance), size=X.shape)

            # bounds
            if self.problem.has_bounds():
                xl, xu = self.problem.bounds()
                X = matrix_conditional_update(X, xl, xu, self.pop.get("X"))

            # recreate the current population without being evaluated
            pop = Population.new(X=X)

        else:

            # recreate the current population without being evaluated
            pop = Population.new(X=X)

            # randomly sample half of the population and reuse half from the previous search
            # when the history information is not enough to build an AR(p) model.

            # randomly sample half of the population
            a = int(self.pop_size / 2)
            pop[:a] = self.initialization.sampling(self.problem, a, random_state=self.random_state)

            # randomly reuse the other half from t Population
            Q = self.pop.get("X")
            b = self.pop_size - a
            idx = self.random_state.choice(np.arange(len(Q)), size=b)
            pop[a:] = Population.new(X=Q[idx])

        return pop

    def center_points_prediction(self, center_points):
        n = len(center_points[0])
        center = np.zeros(n)
        variances = np.zeros(n)
        for i in range(len(center)):
            data = [c[i] for c in center_points]
            model = ARModel(self.p).fit(data)
            predictions = model.predict(data, 1)
            center[i], variances[i] = predictions[0], np.mean(model.resid_ ** 2)
        return center, variances



"""Provide the real-world applications of Dynamic Multi-Objective Optimization."""

import numpy as np

from pydmoo.problems.dyn import DynamicApplProblem


class DSRP(DynamicApplProblem):
    """Dynamic Speed Reducer Problem (DSRP).

    Kurpati, A., Azarm, S., and Wu, J. (2002).
    Constraint handling improvements for multiobjective genetic algorithms.
    Structural and Multidisciplinary Optimization, 23(3), 204–213.
    https://doi.org/10.1007/s00158-002-0178-2
    g9 = 1.9 - x5 + 1.1x7 <= 0 --- [x]
    g10 = f1 - 1300 <= 0 --------- [x]

    Zhang, Z., and Qian, S. (2011).
    Artificial immune system in dynamic environments solving time-varying non-linear constrained multi-objective problems.
    Soft Computing, 15(7), 1333–1349.
    https://doi.org/10.1007/s00500-010-0674-z
    g9 = 1.9 - x5 + 1.1x1 <= 0
    g10 = f1 - 4300 <= 0

    Zhang, Q., He, X., Yang, S., Dong, Y., Song, H., and Jiang, S. (2022).
    Solving dynamic multi-objective problems using polynomial fitting-based prediction algorithm.
    Information Sciences, 610, 868–886.
    https://doi.org/10.1016/j.ins.2022.08.020
    g9 = 1.9 - x5 + 1.1x1 <= 0
    g10 = f1 - 4300 <= 0

    Zou, J., Hou, Z., Jiang, S., Yang, S., Ruan, G., Xia, Y., and Liu, Y. (2025).
    Knowledge transfer with mixture model in dynamic multi-objective optimization.
    IEEE Transactions on Evolutionary Computation, 29(5), 1517–1530.
    https://doi.org/10.1109/TEVC.2025.3566481
    g9 = 1.9 - x5 + 1.1x1 <= 0
    g10 = f1 - 4300 <= 0
    """

    def __init__(self, n_var=7, n_obj=2, nt=10, taut=10, **kwargs):
        super().__init__(
            nt,
            taut,
            n_var=n_var,
            n_obj=n_obj,
            n_ieq_constr=11,
            xl=[2.6, 0.7, 17, 7.3, 7.3, 2.9, 5.0],
            xu=[3.6, 0.8, 28, 8.3, 8.3, 3.9, 5.5],
            **kwargs,
        )

    def _evaluate(self, x, out, *args, **kwargs):
        t = self.time
        x1, x2, x3, x4, x5, x6, x7 = x.T

        f1 = (
            a1_func(t) * x1 * x2**2 * (10 / 3 * x3**2 + a2_func(t) * x3 - a3_func(t))
            - a4_func(t) * x1 * (x6**2 + x7**2)
            + a5_func(t) * (x6**3 + x7**3)
            + a6_func(t) * (x4 * x6**2 + x5 * x7**2)
        )
        f2 = np.sqrt((745.0 * x4 / (x2 * x3)) ** 2 + 1.69e7) / (a7_func(t) * x6**3)

        out["F"] = np.column_stack([f1, f2])

        out["G"] = evaluate_constraints(x, f1)


def a1_func(t):
    a1 = 0.7854
    return a1 + t / 10


def a2_func(t):
    a2 = 14.933
    return a2 + t / 10


def a3_func(t):
    a3 = 43.0934
    return a3 + t / 10


def a4_func(t):
    a4 = 1.508
    return a4 + t / 10


def a5_func(t):
    a5 = 7.477
    return a5 + t / 10


def a6_func(t):
    a6 = 0.7854
    return a6 + t / 10


def a7_func(t):
    a7 = 0.1
    return a7 + (0.25 - 1 / (t + 4))


def evaluate_constraints(x, f1):
    """
    Evaluate all constraint functions for the speed reducer design problem.

    Parameters
    ----------
    x : numpy.ndarray of shape (n_samples, 7)
        Decision variable matrix where each row represents a solution vector
        [x1, x2, x3, x4, x5, x6, x7].
    f1 :
        Objective function value

    Returns
    -------
    g : numpy.ndarray of shape (n_samples, 11)
        Constraint violation matrix. g[:, i] <= 0 indicates the i-th constraint is satisfied.
        Each column corresponds to constraints g1 through g11.
    """
    n_samples = x.shape[0]
    g = np.zeros((n_samples, 11))

    # Extract design variables (0-based indexing in Python)
    x1 = x[:, 0]  # Gear face width
    x2 = x[:, 1]  # Teeth module
    x3 = x[:, 2]  # Number of teeth of pinion
    x4 = x[:, 3]  # Distance between bearings 1
    x5 = x[:, 4]  # Distance between bearings 2
    x6 = x[:, 5]  # Diameter of shaft 1
    x7 = x[:, 6]  # Diameter of shaft 2

    # g1: Geometric constraint related to gear dimensions
    g[:, 0] = 1.0 / (x1 * x2**2 * x3) - 1.0 / 27.0

    # g2: Second geometric constraint for gear design
    g[:, 1] = 1.0 / (x1 * x2**2 * x3**2) - 1.0 / 397.5

    # g3: Stiffness or stress constraint for shaft 1
    # Note: Original text has x5^5 but likely should be x5^3 for symmetry with g3
    g[:, 2] = x4**3 / (x2 * x3 * x6**4) - 1.0 / 1.93

    # g4: Stiffness or stress constraint for shaft 2
    g[:, 3] = x5**3 / (x2 * x3 * x7**4) - 1.0 / 1.93

    # g5: Constraint on gear size (teeth module x number of teeth)
    g[:, 4] = x2 * x3 - 40.0

    # g6: Upper bound constraint on face width to module ratio
    g[:, 5] = x1 / x2 - 12.0

    # g7: Lower bound constraint on face width to module ratio (rearranged form)
    g[:, 6] = 5.0 - x1 / x2

    # g8: Relationship between bearing distance 1 and shaft 1 diameter
    g[:, 7] = 1.9 - x4 + 1.5 * x6

    # g9: Relationship between bearing distance 2 and gear face width
    g[:, 8] = 1.9 - x5 + 1.1 * x1

    # g10: Upper bound constraint on weight objective function f1
    g[:, 9] = f1 - 4300.0

    # g11: Stress constraint for shaft 2
    g[:, 10] = np.sqrt((745.0 * x5 / (x2 * x3)) ** 2 + 1.515e8) / (0.1 * x7**3) - 1100.0

    return g



"""
Includes modified code from [pymoo](https://github.com/anyoptimization/pymoo).

Sources:
    - [algorithm.py](https://github.com/anyoptimization/pymoo/blob/main/pymoo/core/algorithm.py).

Licensed under the Apache License, Version 2.0. Original copyright and license terms are preserved.
"""

import copy
import time

import numpy as np
from pymoo.core.callback import Callback
from pymoo.core.evaluator import Evaluator
from pymoo.core.meta import Meta
from pymoo.core.population import Population
from pymoo.core.result import Result
from pymoo.functions import FunctionLoader
from pymoo.termination.default import DefaultMultiObjectiveTermination, DefaultSingleObjectiveTermination
from pymoo.util.display.display import Display
from pymoo.util.misc import termination_from_tuple
from pymoo.util.optimum import filter_optimum


class Algorithm:

    def __init__(self,
                 termination=None,
                 output=None,
                 display=None,
                 callback=None,
                 archive=None,
                 return_least_infeasible=False,
                 save_history=False,
                 verbose=False,
                 seed=None,
                 evaluator=None,
                 **kwargs):

        super().__init__()

        # prints the compile warning if enabled
        FunctionLoader.get_instance()

        # the problem to be solved (will be set later on)
        self.problem = None

        # the termination criterion to be used by the algorithm - might be specific for an algorithm
        self.termination = termination

        # the text that should be printed during the algorithm run
        self.output = output

        # an archive kept during algorithm execution (not always the same as optimum)
        self.archive = archive

        # the form of display shown during algorithm execution
        self.display = display

        # callback to be executed each generation
        if callback is None:
            callback = Callback()
        self.callback = callback

        # whether the algorithm should finally return the least infeasible solution if no feasible found
        self.return_least_infeasible = return_least_infeasible

        # whether the history should be saved or not
        self.save_history = save_history

        # whether the algorithm should print output in this run or not
        self.verbose = verbose

        # the random seed that was used
        self.seed = seed
        self.random_state = None

        # the function evaluator object (can be used to inject code)
        if evaluator is None:
            evaluator = Evaluator()
        self.evaluator = evaluator

        # the history object which contains the list
        self.history = list()

        # the current solutions stored - here considered as population
        self.pop = None

        # a placeholder object for implementation to store solutions in each iteration
        self.off = None

        # the optimum found by the algorithm
        self.opt = None

        # the current number of generation or iteration
        self.n_iter = None

        # can be used to store additional data in submodules
        self.data = {}

        # if the initialized method has been called before or not
        self.is_initialized = False

        # the time when the algorithm has been setup for the first time
        self.start_time = None

    def setup(self, problem, verbose=False, progress=False, **kwargs):

        # the problem to be solved by the algorithm
        self.problem = problem

        # clone the output object if it exists to avoid state pollution between runs
        if self.output is not None:
            self.output = copy.deepcopy(self.output)

        # set all the provided options to this method
        for key, value in kwargs.items():
            self.__dict__[key] = value

        # set random state
        self.random_state = np.random.default_rng(self.seed)

        # make sure that some type of termination criterion is set
        if self.termination is None:
            self.termination = default_termination(problem)
        else:
            self.termination = termination_from_tuple(self.termination)

        # set up the display during the algorithm execution
        if self.display is None:
            self.display = Display(self.output, verbose=verbose, progress=progress)

        # finally call the function that can be overwritten by the actual algorithm
        self._setup(problem, **kwargs)

        return self

    def run(self):
        while self.has_next():
            self.next()
        return self.result()

    def has_next(self):
        return not self.termination.has_terminated()

    def finalize(self):

        # finalize the display output in the end of the run
        self.display.finalize()

        return self._finalize()

    def next(self):

        # get the infill solutions
        infills = self.infill()

        # call the advance with them after evaluation
        if infills is not None:
            self.evaluator.eval(self.problem, infills, algorithm=self)
            self.advance(infills=infills)

        # if the algorithm does not follow the infill-advance scheme just call advance
        else:
            self.advance()

    def _initialize(self):

        # the time starts whenever this method is called
        self.start_time = time.time()

        # set the attribute for the optimization method to start
        self.n_iter = 1
        self.pop = Population.empty()
        self.opt = None

    def infill(self):
        if self.problem is None:
            raise Exception("Please call `setup(problem)` before calling next().")

        # the first time next is called simply initial the algorithm - makes the interface cleaner
        if not self.is_initialized:

            # hook mostly used by the class to happen before even to initialize
            self._initialize()

            # execute the initialization infill of the algorithm
            infills = self._initialize_infill()

        else:
            # request the infill solutions if the algorithm has implemented it
            infills = self._infill()

        # set the current generation to the offsprings
        if infills is not None:
            infills.set("n_gen", self.n_iter)
            infills.set("n_iter", self.n_iter)

        return infills

    def advance(self, infills=None, **kwargs):

        # if infills have been provided set them as offsprings and feed them into advance
        self.off = infills

        # if the algorithm has not been already initialized
        if not self.is_initialized:

            # set the generation counter to 1
            self.n_iter = 1

            # assign the population to the algorithm
            self.pop = infills

            # do what is necessary after the initialization
            self._initialize_advance(infills=infills, **kwargs)

            # set this algorithm to be initialized
            self.is_initialized = True

            # always advance to the next iteration after initialization
            self._post_advance()

        else:

            # call the implementation of the advance method - if the infill is not None
            val = self._advance(infills=infills, **kwargs)

            # always advance to the next iteration - except if the algorithm returns False
            if val is None or val:
                self._post_advance()

        # if the algorithm has terminated, then do the finalization steps and return the result
        if self.termination.has_terminated():
            self.finalize()
            ret = self.result()

        # otherwise just increase the iteration counter for the next step and return the current optimum
        else:
            ret = self.opt

        # add the infill solutions to an archive
        if self.archive is not None and infills is not None:
            self.archive = self.archive.add(infills)

        return ret

    def result(self):
        res = Result()

        # store the time when the algorithm as finished
        res.start_time = self.start_time
        res.end_time = time.time()
        res.exec_time = res.end_time - res.start_time

        res.pop = self.pop
        res.archive = self.archive
        res.data = self.data

        # get the optimal solution found
        opt = self.opt
        if opt is None or len(opt) == 0:
            opt = None

        # if no feasible solution has been found
        elif not np.any(opt.get("FEAS")):
            if self.return_least_infeasible:
                opt = filter_optimum(opt, least_infeasible=True)
            else:
                opt = None
        res.opt = opt

        # if optimum is set to none to not report anything
        if res.opt is None:
            X, F, CV, G, H = None, None, None, None, None

        # otherwise get the values from the population
        else:
            X, F, CV, G, H = self.opt.get("X", "F", "CV", "G", "H")

            # if single-objective problem and only one solution was found - create a 1d array
            if self.problem.n_obj == 1 and len(X) == 1:
                X, F, CV, G, H = X[0], F[0], CV[0], G[0], H[0]

        # set all the individual values
        res.X, res.F, res.CV, res.G, res.H = X, F, CV, G, H

        # create the result object
        res.problem = self.problem
        res.history = self.history

        return res

    def ask(self):
        return self.infill()

    def tell(self, *args, **kwargs):
        return self.advance(*args, **kwargs)

    def _set_optimum(self):
        self.opt = filter_optimum(self.pop, least_infeasible=True)

    def _post_advance(self):

        # update the current optimum of the algorithm
        self._set_optimum()

        # update the current termination condition of the algorithm
        self.termination.update(self)

        # display the output if defined by the algorithm
        self.display(self)

        if self.save_history:
            _hist, _callback, _display = self.history, self.callback, self.display

            self.history, self.callback, self.display = None, None, None
            obj = copy.deepcopy(self)

            self.history, self.callback, self.display = _hist, _callback, _display
            self.history.append(obj)

        # if a callback function is provided it is called after each iteration
        self.callback(self)

        self.n_iter += 1

    # =========================================================================================================
    # TO BE OVERWRITTEN
    # =========================================================================================================

    def _setup(self, problem, **kwargs):
        pass

    def _initialize_infill(self):
        pass

    def _initialize_advance(self, infills=None, **kwargs):
        pass

    def _infill(self):
        pass

    def _advance(self, infills=None, **kwargs):
        pass

    def _finalize(self):
        pass

    # =========================================================================================================
    # CONVENIENCE
    # =========================================================================================================

    @property
    def n_gen(self):
        return self.n_iter

    @n_gen.setter
    def n_gen(self, value):
        self.n_iter = value


class LoopwiseAlgorithm(Algorithm):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.generator = None
        self.state = None

    def _next(self):
        pass

    def _infill(self):
        if self.state is None:
            self._advance()
        return self.state

    def _advance(self, infills=None, **kwargs):
        if self.generator is None:
            self.generator = self._next()
        try:
            self.state = self.generator.send(infills)
        except StopIteration:
            self.generator = None
            self.state = None
            return True

        return False


def default_termination(problem):
    if problem.n_obj > 1:
        termination = DefaultMultiObjectiveTermination()
    else:
        termination = DefaultSingleObjectiveTermination()
    return termination


class MetaAlgorithm(Meta):
    """
    An algorithm wrapper that combines Algorithm's functionality with Meta's delegation behavior.
    Uses Meta to provide transparent proxying with the ability to override specific methods.
    """

    def __init__(self, algorithm, copy=True, **kwargs):
        # If the algorithm is already a Meta object, don't copy to avoid deepcopy issues with nested proxies
        if isinstance(algorithm, Meta):
            copy = False

        # Initialize Meta
        super().__init__(algorithm, copy=copy)

        # Pass any additional kwargs to the wrapped algorithm if needed
        for key, value in kwargs.items():
            setattr(self, key, value)



from pydmoo.problems.dynamic.cec2015 import FDA4, FDA5
from pydmoo.problems.dynamic.df import DF1, DF2, DF3, DF4, DF5, DF6, DF7, DF8, DF9, DF10, DF11, DF12, DF13, DF14
from pydmoo.problems.dynamic.gts import (
    GTS1,
    GTS1_2,
    GTS1_3,
    GTS2,
    GTS2_2,
    GTS2_3,
    GTS3,
    GTS3_2,
    GTS3_3,
    GTS4,
    GTS4_2,
    GTS4_3,
    GTS5,
    GTS5_2,
    GTS5_3,
    GTS6,
    GTS6_2,
    GTS6_3,
    GTS7,
    GTS7_2,
    GTS7_3,
    GTS8,
    GTS8_2,
    GTS8_3,
    GTS9,
    GTS9_2,
    GTS9_3,
    GTS10,
    GTS10_2,
    GTS10_3,
    GTS11,
    GTS11_2,
    GTS11_3,
)
from pydmoo.problems.real_world.dsrp import DSRP
from pydmoo.problems.real_world.dwbdp import DWBDP

PROBLEM = {
    "fda4": FDA4,
    "fda5": FDA5,
    "gts1": GTS1,
    "gts2": GTS2,
    "gts3": GTS3,
    "gts4": GTS4,
    "gts5": GTS5,
    "gts6": GTS6,
    "gts7": GTS7,
    "gts8": GTS8,
    "gts9": GTS9,
    "gts10": GTS10,
    "gts11": GTS11,
    "gts1_2": GTS1_2,
    "gts2_2": GTS2_2,
    "gts3_2": GTS3_2,
    "gts4_2": GTS4_2,
    "gts5_2": GTS5_2,
    "gts6_2": GTS6_2,
    "gts7_2": GTS7_2,
    "gts8_2": GTS8_2,
    "gts9_2": GTS9_2,
    "gts10_2": GTS10_2,
    "gts11_2": GTS11_2,
    "gts1_3": GTS1_3,
    "gts2_3": GTS2_3,
    "gts3_3": GTS3_3,
    "gts4_3": GTS4_3,
    "gts5_3": GTS5_3,
    "gts6_3": GTS6_3,
    "gts7_3": GTS7_3,
    "gts8_3": GTS8_3,
    "gts9_3": GTS9_3,
    "gts10_3": GTS10_3,
    "gts11_3": GTS11_3,
    "df1": DF1,
    "df2": DF2,
    "df3": DF3,
    "df4": DF4,
    "df5": DF5,
    "df6": DF6,
    "df7": DF7,
    "df8": DF8,
    "df9": DF9,
    "df10": DF10,
    "df11": DF11,
    "df12": DF12,
    "df13": DF13,
    "df14": DF14,
    "dsrp": DSRP,
    "dwbdp": DWBDP,
}


def get_problem(name, *args, **kwargs):
    name = name.lower()

    if name.lower() not in PROBLEM:
        raise Exception("Problem not found.")

    return PROBLEM[name.lower()](*args, **kwargs)



import numpy as np


def univariate_gaussian_sample(mean, std, n_samples=100, random_state=None):
    """
    Generate samples from a 1-dimensional Gaussian distribution.

    Parameters
    ----------
    mean : array_like, shape (n_features,)
        Mean values of the Gaussian distribution for each feature.
    std : array_like, shape (n_features,)
        Standard deviation of the Gaussian distribution for each feature.
    n_samples : int, optional
        Number of samples to generate (default=100).

    Returns
    -------
    ndarray, shape (n_samples, n_features)
        Random samples from the specified Gaussian distribution.

    Notes
    -----
    This function generates independent 1D Gaussian samples for each feature.
    """
    return random_state.normal(mean, std, size=(n_samples, len(mean)))  # np.random


def multivariate_gaussian_sample(mean, cov, n_samples=1, random_state=None):
    """
    Generate samples from a multivariate Gaussian distribution.

    Parameters
    ----------
    mean : array_like, shape (n_features,)
        Mean vector of the distribution.
    cov : array_like, shape (n_features, n_features)
        Covariance matrix of the distribution.
    n_samples : int, optional
        Number of samples to generate (default=1).

    Returns
    -------
    ndarray, shape (n_samples, n_features)
        Random samples from the multivariate Gaussian distribution.

    Raises
    ------
    ValueError
        If the covariance matrix is not positive-semidefinite.

    Notes
    -----
    Uses numpy.random.multivariate_normal for sampling.
    """
    return random_state.multivariate_normal(mean, cov, size=n_samples)  # np.random



import numpy as np


def closed_form_solution(X, Y):
    """
    Compute the least squares solution B = (YᵀY)⁻¹(YᵀX) for the linear system X ≈ YB.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        Target matrix containing dependent variables.
    Y : ndarray, shape (m, p)
        Design matrix containing independent variables.

    Returns
    -------
    B : ndarray, shape (p, n)
        Coefficient matrix that minimizes the Frobenius norm of (X - YB).

    Raises
    ------
    LinAlgError
        If YᵀY is singular and cannot be inverted, falls back to pseudo-inverse.

    Notes
    -----
    1. Solves the ordinary least squares problem in closed form.
    2. Automatically handles singular matrices by using pseudo-inverse when necessary.
    3. For numerical stability, consider using np.linalg.lstsq() in production code.

    """
    # Method 1
    # # Compute Y transpose multiplied by X: YᵀX (p x n)
    # YTX = np.dot(Y.T, X)

    # # Compute Y transpose multiplied by Y: YᵀY (p x p)
    # YTY = np.dot(Y.T, Y)

    # # Compute inverse of YᵀY with fallback to pseudo-inverse
    # try:
    #     YTY_inv = np.linalg.inv(YTY)
    # except np.linalg.LinAlgError:
    #     YTY_inv = np.linalg.pinv(YTY)

    # # Compute final solution: B = (YᵀY)⁻¹(YᵀX)
    # B = np.dot(YTY_inv, YTX)

    # Method 2
    B, residuals, rank, s = np.linalg.lstsq(Y, X, rcond=None)

    # Method 3
    # from scipy.linalg import lstsq
    # B = lstsq(Y, X, cond=None)[0]

    return B



import copy

import numpy as np
from pymoo.core.population import Population
from scipy.optimize import Bounds, minimize

from pydmoo.algorithms.base.dmoo.dmoeadde import DMOEADDE
from pydmoo.core.transfer import TCA


class MOEADDETr(DMOEADDE):
    """Transfer learning (Tr).

    Transfer Learning-based Initial Population Generator (Tr-IPG)

    References
    ----------
    Jiang, M., Huang, Z., Qiu, L., Huang, W., and Yen, G. G. (2018).
    Transfer learning-based dynamic multiobjective optimization algorithms.
    IEEE Transactions on Evolutionary Computation, 22(4), 501–514.
    https://doi.org/10.1109/TEVC.2017.2771451
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.ndim_ls = 20  # the dimension of latent space
        self.mu = 0.5

        #
        self._maxiter = max(self.pop_size, 100)  # default is 1000

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X, F = pop.get("X", "F")

        last_time = self.data.get("last_time", 0)
        self.data["last_time"] = self.problem.time

        # source domain
        problem_ = copy.deepcopy(self.problem)
        problem_.time = last_time
        pop_s = self.initialization.sampling(problem_, self.pop_size, random_state=self.random_state)
        pop_s = self.evaluator.eval(problem_, pop_s)

        # target domain
        pop_t = self.initialization.sampling(self.problem, self.pop_size, random_state=self.random_state)
        pop_t = self.evaluator.eval(self.problem, pop_t)

        # Algorithm 1: TCA
        model = TCA(dim=self.ndim_ls, mu=self.mu)
        model.fit(pop_s.get("F"), pop_t.get("F"))

        # Remark3
        particles_latent_saace = model.transform(self.opt.get("F"))

        def dist_px(p, x, xl, xu):
            x = np.clip(x, xl, xu)
            pop_temp = Population.new(X=[x])
            pop_temp = self.evaluator.eval(self.problem, pop_temp)
            F = pop_temp.get("F")
            return np.sum((model.transform(F) - p) ** 2)

        X_ = []
        xl, xu = self.problem.bounds()
        for particle in particles_latent_saace:
            start = self.initialization.sampling(self.problem, 1, random_state=self.random_state).get("X")[0]
            start = np.clip(start, xl, xu)

            try:
                res = minimize(
                    lambda x: dist_px(particle, x, xl, xu),
                    start,
                    bounds=Bounds(xl, xu),
                    method="trust-constr",  # SLSQP, trust-constr, L-BFGS-B; In this paper, we use the interior point algorithm to solve the problem.
                    options={
                        "maxiter": self._maxiter,
                    },
                )
                x_opt = np.clip(res.x, xl, xu)
                X_.append(x_opt)

            except Exception as e:
                random_point = self.initialization.sampling(self.problem, 1, random_state=self.random_state).get("X")[0]
                X_.append(np.clip(random_point, xl, xu))

        # bounds
        X_ = np.array(X_)
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X_ = np.clip(X_, xl, xu)  # not provided in the original reference literature

        # recreate the current population without being evaluated
        # merge
        pop = Population.merge(pop_t, Population.new(X=X_))

        return pop



from pymoo.core.population import Population

from pydmoo.algorithms.modern.moead_imkt import MOEADIMKT
from pydmoo.core.bounds import clip_and_randomize
from pydmoo.core.predictions import igp_based_predictor
from pydmoo.core.sample_gaussian import univariate_gaussian_sample


class MOEADIMKTIGP(MOEADIMKT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5

        self.delta_s = 0.01
        self.sigma_n = 0.01
        self.sigma_n_2 = self.sigma_n ** 2

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # sample self.pop_size individuals in decision space
        samples_old = self.sampling_new_pop()

        # select self.pop_size/2 individuals with better convergence and diversity
        samples = samples_old[:int(len(samples_old)/2)]

        # knowledge in objective space
        means_stds, mean, std = self._in_decision_or_objective_space_1d(samples, "objective_space")
        mean_new, std_new = self._select_means_stds(means_stds, mean, std)

        # sample self.pop_size individuals in objective space
        F = univariate_gaussian_sample(mean_new, std_new, self.pop_size, random_state=self.random_state)

        # TODO
        # inverse mapping
        X = igp_based_predictor(samples.get("X"), samples.get("F"), F, self.sigma_n_2)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = clip_and_randomize(X, xl, xu, random_state=self.random_state)

        # merge
        pop = Population.merge(samples_old, Population.new(X=X))

        return pop


class MOEADIMKTIGP0(MOEADIMKTIGP):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import numpy as np
from pymoo.core.population import Population

from pydmoo.algorithms.modern.nsga2_imkt_n import NSGA2IMKTN
from pydmoo.algorithms.utils.utils import make_semidefinite, reconstruct_covariance_from_triu
from pydmoo.core.bounds import clip_and_randomize
from pydmoo.core.inverse import closed_form_solution
from pydmoo.core.lstm.lstm import LSTMpredictor
from pydmoo.core.sample_gaussian import multivariate_gaussian_sample


class NSGA2IMNLSTM(NSGA2IMKTN):
    """Inverse Modeling with LSTM (IMNLSTM).

    Inverse Modeling for Dynamic Multiobjective Optimization with Knowledge Transfer In objective Space.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5

        self._n_timesteps = 10
        self._sequence_length = 5  # Use 5 historical time steps to predict next step
        self._incremental_learning = False

    def _setup(self, problem, **kwargs):
        super()._setup(problem, **kwargs)

        # Must be here
        self._lstm = LSTMpredictor(
            self._sequence_length,
            hidden_dim=64,
            num_layers=1,
            epochs=50,
            batch_size=32,
            lr=0.001,
            device="cpu",  # for fair comparison
            patience=5,
            seed=self.seed,
            model_type="lstm",
            incremental_learning=self._incremental_learning,
        )

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # sample self.pop_size individuals in decision space
        samples_old = self.sampling_new_pop()

        # select self.pop_size/2 individuals with better convergence and diversity
        samples = samples_old[:int(len(samples_old)/2)]

        # knowledge in objective space
        means_covs, mean, cov = self._in_decision_or_objective_space_nd(samples, "objective_space")

        # Check if sufficient historical data is available for LSTM prediction
        if len(means_covs) > self._n_timesteps:
            # Update pool
            self.data["means_covs"] = means_covs[self._n_timesteps:]

            # Prepare time series data from historical means and covariance matrices
            time_series_data = prepare_data_mean_cov(self._n_timesteps, means_covs)

            # Initialize predictor and generate prediction for next time step
            next_prediction = self._lstm.convert_train_predict(time_series_data)

            # Convert prediction tensor to numpy array for further processing
            next_prediction = next_prediction.numpy()

            # Split prediction into mean and covariance components
            # First n_obj elements represent the mean vector, Remaining elements represent the flattened covariance matrix values
            mean_new, cov_new_ = next_prediction[:self.problem.n_obj], next_prediction[self.problem.n_obj:]
            cov_new = reconstruct_covariance_from_triu(cov_new_, len(mean_new))
            cov_new = make_semidefinite(cov_new)

        else:
            mean_new, cov_new = self._select_means_covs(means_covs, mean, cov)

        # sample self.pop_size individuals in objective space
        F = multivariate_gaussian_sample(mean_new, cov_new, self.pop_size, random_state=self.random_state)

        # TODO
        # inverse mapping
        # X = FB
        B = closed_form_solution(samples.get("X"), samples.get("F"))

        # X = FB
        X = np.dot(F, B)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = clip_and_randomize(X, xl, xu, random_state=self.random_state)

        # merge
        pop = Population.merge(samples_old, Population.new(X=X))

        return pop


def prepare_data_mean_cov(n_timesteps, means_covs):
    """Prepare time series data from means and covariance matrices.

    This function converts a sequence of mean vectors and covariance matrices
    into a time series format suitable for LSTM training. It extracts the mean
    values and the upper triangular elements of the covariance matrix (including
    diagonal) to create feature vectors for each time step.

    Parameters
    ----------
    means_covs : list of tuples
        List containing (mean, cov, n_iter) pairs for each time step, where:
        - mean: 1D numpy array of mean values
        - cov: 2D numpy array representing covariance matrix
        - n_iter: number of iterations (not used in feature extraction)

    Returns
    -------
    time_series_data : list
        Combined feature data with shape (n_timesteps, n_features)
        Each row represents a time step containing:
        [mean_1, mean_2, ..., mean_n, cov_11, cov_12, ..., cov_nn]
    """
    # Create time series data
    time_series_data = []  # shape: (n_timesteps, n_features)

    # Process only the most recent n_timesteps
    for m, c, _ in means_covs[-n_timesteps:]:
        # Extract upper triangular indices including diagonal
        # np.triu_indices_from(c) returns row and column indices of upper triangle
        rows, cols = np.triu_indices_from(c)

        # Combine mean vector and upper triangular covariance elements
        # [*m] unpacks all mean values
        # [*c[rows, cols]] unpacks all upper triangular covariance elements
        feature_vector = [*m, *c[rows, cols]]
        time_series_data.append(feature_vector)

    return time_series_data


class NSGA2IMNiLSTM(NSGA2IMNLSTM):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5
        self._n_timesteps = 10
        self._sequence_length = 5  # Use 5 historical time steps to predict next step
        self._incremental_learning = True



import numpy as np
from pymoo.core.population import Population

from pydmoo.algorithms.modern.moead_imkt import MOEADIMKT
from pydmoo.core.bounds import clip_and_randomize
from pydmoo.core.distance import norm_mean_frobenius_distance
from pydmoo.core.inverse import closed_form_solution
from pydmoo.core.sample_gaussian import multivariate_gaussian_sample


class MOEADIMKTN(MOEADIMKT):
    """Inverse Modeling with Knowledge Transfer.

    Inverse Modeling for Dynamic Multiobjective Optimization with Knowledge Transfer In objective Space.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_pool = 10
        self.denominator = 0.5

    def _response_mechanism(self) -> Population:
        """Response mechanism."""
        pop = self.pop
        X = pop.get("X")

        # recreate the current population without being evaluated
        pop = Population.new(X=X)

        # sample self.pop_size individuals in decision space
        samples_old = self.sampling_new_pop()

        # select self.pop_size/2 individuals with better convergence and diversity
        samples = samples_old[:int(len(samples_old)/2)]

        # knowledge in objective space
        means_stds, mean, cov = self._in_decision_or_objective_space_nd(samples, "objective_space")
        mean_new, cov_new = self._select_means_covs(means_stds, mean, cov)

        # sample self.pop_size individuals in objective space
        F = multivariate_gaussian_sample(mean_new, cov_new, self.pop_size, random_state=self.random_state)

        # TODO
        # inverse mapping
        # X = FB
        B = closed_form_solution(samples.get("X"), samples.get("F"))

        # X = FB
        X = np.dot(F, B)

        # bounds
        if self.problem.has_bounds():
            xl, xu = self.problem.bounds()
            X = clip_and_randomize(X, xl, xu, random_state=self.random_state)

        # merge
        pop = Population.merge(samples_old, Population.new(X=X))

        return pop

    def _in_decision_or_objective_space_nd(self, samples, decision_or_objective="decision_space"):
        # decision space or objective space
        flag = "X" if decision_or_objective == "decision_space" else "F"

        means_covs = self.data.get("means_covs", [])

        flag_value = self.opt.get(flag)
        if len(flag_value) <= 1:
            flag_value = self.pop.get(flag)
            flag_value = flag_value[:2]

        m, c = np.mean(flag_value, axis=0), np.cov(flag_value.T)
        means_covs.append((m, 0.5 * (c.T + c), self.n_iter - 1))  # 1-based
        self.data["means_covs"] = means_covs

        flag_value = samples.get(flag)
        mean, cov = np.mean(flag_value, axis=0), np.cov(flag_value.T)
        return means_covs, mean, 0.5 * (cov.T + cov)

    def _select_means_covs(self, means_covs, mean_new, cov_new):
        # Unpack means and stds
        means = np.array([m[0] for m in means_covs])
        covs = np.array([m[1] for m in means_covs])

        # Calculate distances
        distances = np.array([
            norm_mean_frobenius_distance(mean, cov, mean_new, cov_new) for mean, cov in zip(means, covs)
        ])

        # Get top K closest
        top_k_idx = np.argsort(distances)[:self.size_pool]
        top_k_dist = distances[top_k_idx]
        top_k_means = means[top_k_idx]
        top_k_covs = covs[top_k_idx]

        # Update pool
        self._update_means_covs_pool(means_covs, top_k_idx)

        # Calculate weights
        weights = 1 / (top_k_dist + 1e-8)  # Add small epsilon to avoid division by zero
        weights = weights / (np.sum(weights) + self.denominator)

        # Weighted combination
        mean_new = (1 - np.sum(weights)) * mean_new + np.sum(weights[:, None] * top_k_means, axis=0)
        cov_new = (1 - np.sum(weights)) * cov_new + np.sum(weights[:, None, None] * top_k_covs, axis=0)

        # Symmetric matrix
        cov_new = 0.5 * (cov_new.T + cov_new)
        return mean_new, cov_new

    def _update_means_covs_pool(self, means_covs, top_k_idx) -> None:
        self.data["means_covs"] = [means_covs[i] for i in top_k_idx]
        return None


class MOEADIMKTN0(MOEADIMKTN):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import numpy as np

from pydmoo.problems.dyn import DynamicApplProblem


class DWBDP(DynamicApplProblem):
    r"""Dynamic Welded Beam Design Problem (DWBDP).

    A multi-objective optimization problem with time-varying constraints and objectives
    that minimizes both manufacturing cost and beam deflection under dynamic loading conditions.

    Objectives
    ----------
    f1 : float
        Manufacturing cost ($), comprising setup cost, welding labor cost, and material cost
    f2 : float
        Beam deflection (inch) under applied load

    Variables
    ---------
    x : ndarray, shape (4,)
        x[0] : weld thickness h (inch)
        x[1] : weld length l (inch)
        x[2] : beam height t (inch)
        x[3] : beam thickness b (inch)

    Constraints
    -----------
    g1 : float
        Shear stress constraint ($\tau <= 13,600$)
    g2 : float
        Bending stress constraint ($\sigma ≤ 30,000$)
    g3 : float
        Geometric constraint ($h <= b$)
    g4 : float
        Buckling constraint ($P <= P_critical$)

    Notes
    -----
    It should be noted that in the formulation [1], the variable used in the expression
    for the resultant shear stress $\tau$ is $x_2$, rather than $x_3$ as may appear in certain
    alternative formulations. Readers are advised to verify the specific variable
    definitions and indexing used in the respective reference to ensure consistency
    in implementation.

    The problem incorporates temporal variations through time-dependent loading
    conditions P(t), making it suitable for testing dynamic multi-objective
    optimization algorithms.

    References
    ----------
    Oyama, A., Shimoyama, K., and Fujii, K. (2005).
    New constraint-handling method for multi-objective multi-constraint evolutionary optimization and its application to space plane design.
    Evolutionary and Deterministic Methods for Design, Optimization and Control with Applications to Industrial and Societal Problems.
    https://ladse.eng.isas.jaxa.jp/papers/eurogen2005.pdf

    Zhang, Z., and Qian, S. (2011).
    Artificial immune system in dynamic environments solving time-varying non-linear constrained multi-objective problems.
    Soft Computing, 15(7), 1333–1349.
    https://doi.org/10.1007/s00500-010-0674-z

    Zhang, Q., He, X., Yang, S., Dong, Y., Song, H., and Jiang, S. (2022).
    Solving dynamic multi-objective problems using polynomial fitting-based prediction algorithm.
    Information Sciences, 610, 868–886.
    https://doi.org/10.1016/j.ins.2022.08.020

    https://www.mathworks.com/help/gads/multiobjective-optimization-welded-beam.html

    https://github.com/Amir-M-Vahedi/Welded-Beam-Design-Optimization/blob/main/Welded_Beam_Design_Project_D3.py
    """

    def __init__(self, n_var=4, n_obj=2, nt=1, taut=10, **kwargs):
        # COMPULSORY
        nt = 1

        super().__init__(
            nt, taut, n_var=n_var, n_obj=n_obj, n_ieq_constr=4, xl=[55, 75, 1000, 2], xu=[80, 110, 3000, 20], **kwargs
        )

        # Material properties and constants
        self.E = 3.0e7  # Young's modulus (psi)
        self.G = 1.2e7  # Shear modulus (psi)
        self.L = 14.0  # Beam length (inch)

        # Load cases: (time, load in lb)
        self.load_cases = {0: 10000, 1: 80000, 2: 6000, 3: 3000}

        # [55, 80],      # x1: weld thickness bounds
        # [75, 110],     # x2: weld length bounds
        # [1000, 3000],  # x3: beam height bounds
        # [2, 20],       # x4: beam thickness bounds

        #
        # lb = [55, 75, 1000, 2]
        # ub = [80, 110, 3000, 20]

        #
        # lb = [0.125, 0.1, 0.1, 0.125]
        # ub = [5, 10, 10 ,5]

    def _evaluate(self, x, out, *args, **kwargs):
        t = self.time
        x1, x2, x3, x4 = x.T
        P = self.load_cases[int(t % 4)]

        f1 = f1 = 1.10471 * x1**2 * x2 + 0.04811 * x3 * x4 * (14 + x2)

        f2 = (4 * P * self.L**3) / (self.E * x3**3 * x4)

        out["F"] = np.column_stack([f1, f2])

        out["G"] = self.evaluate_constraints(x, P)

    def calculate_auxiliary_variables(self, x, P) -> tuple:
        """Calculate intermediate variables needed for constraint evaluation.

        Parameters
        ----------
        x : np.ndarray, shape (4,)
            Design variables [x1, x2, x3, x4]
        P : float
            Applied load (lb)

        Returns
        -------
        tuple : (tau, pc)
            tau : float - Resultant shear stress (psi)
            pc : float - Critical buckling load (lb)
        """
        x1, x2, x3, x4 = x.T

        # Radius to neutral axis
        R = np.sqrt(x2**2 / 4 + ((x1 + x3) / 2) ** 2)

        # Polar moment of inertia
        J = 2 * np.sqrt(2) * x1 * x2 * (x2**2 / 12 + ((x1 + x3) / 2) ** 2)

        # Bending moment
        M = P * (self.L + x2 / 2)

        # Primary and secondary shear stresses
        tau1 = P / (np.sqrt(2) * x1 * x2)
        tau2 = (M * R) / J

        # Resultant shear stress
        tau = np.sqrt(tau1**2 + 2 * tau1 * tau2 * (x2 / (2 * R)) + tau2**2)

        # Critical buckling load
        pc_numerator = 4.013 * self.E * x3 * x4**3
        pc_denominator = 6 * self.L**2
        pc_adjustment = 1 - (x3 / (2 * self.L)) * np.sqrt(self.E / (4 * self.G))
        pc = (pc_numerator / pc_denominator) * pc_adjustment

        return tau, pc

    def evaluate_constraints(self, x, P):
        """Evaluate all constraint functions.

        Parameters
        ----------
        x : np.ndarray, shape (4,)
            Design variables [x1, x2, x3, x4]

        Returns
        -------
        np.ndarray
            Constraint values (should be <= 0 for feasibility)
        """
        n_samples = x.shape[0]
        g = np.zeros((n_samples, 4))

        x1, _, x3, x4 = x.T

        # Evaluate constraints for specific load case
        tau, pc = self.calculate_auxiliary_variables(x, P)

        # g1: Shear stress constraint (tau <= 13,600 psi)
        g[:, 0] = tau - 13600

        # g2: Bending stress constraint (sigma <= 30,000 psi)
        g[:, 1] = (6 * P * self.L) / (x4 * x3**2) - 30000

        # g3: Geometric constraint (x1 <= x4)
        g[:, 2] = x1 - x4

        # g4: Buckling constraint (P <= pc)
        g[:, 3] = P - pc

        return g



"""
Includes modified code from [pymoo](https://github.com/anyoptimization/pymoo).

Sources:

- [cec2015.py](https://github.com/anyoptimization/pymoo/blob/main/pymoo/problems/dynamic/cec2015.py)

Licensed under the Apache License, Version 2.0. Original copyright and license terms are preserved.

Add the method `_calc_pareto_front` for FDA4 and FDA5.
"""

from math import cos, fabs, floor, pi, sin, sqrt

import numpy as np

from pydmoo.problems.dyn import DynamicTestProblem
from pydmoo.problems.dynamic.df import get_PF


class DynamicCEC2015(DynamicTestProblem):

    def __init__(self, n_var=10, nt=10, taut=20, n_obj=2, xl=0.0, xu=1.0, vtype=float, **kwargs):
        super().__init__(nt, taut, n_var=n_var, n_obj=n_obj, xl=xl, xu=xu, vtype=vtype, **kwargs)


class FDA2DEB(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import fda2_deb as f
        out["F"] = np.array([f(x, t) for x in X])


class FDA4(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, n_obj=3, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import FDA4 as f
        out["F"] = np.array([f(x, t) for x in X])

    # Added by DynOpt
    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')

        f1 = np.cos(x1 * pi / 2) * np.cos(x2 * pi / 2)
        f2 = np.cos(x1 * pi / 2) * np.sin(x2 * pi / 2)
        f3 = np.sin(x1 * pi / 2)

        h = get_PF(np.array([f1, f2, f3]), False)
        return h


class FDA5(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, n_obj=3, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import FDA5 as f
        out["F"] = np.array([f(x, t) for x in X])

    # Added by DynOpt
    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')

        G = fabs(sin(0.5 * pi * self.time))
        g = G
        F = 1 + 100 * pow(sin(0.5 * pi * self.time), 4)

        x1 = np.power(x1, F)
        x2 = np.power(x2, F)

        f1 = (1 + g) * np.cos(x1 * pi / 2) * np.cos(x2 * pi / 2)
        f2 = (1 + g) * np.cos(x1 * pi / 2) * np.sin(x2 * pi / 2)
        f3 = (1 + g) * np.sin(x1 * pi / 2)

        h = get_PF(np.array([f1, f2, f3]), True)
        return h


class DIMP2(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import DIMP2 as f
        out["F"] = np.array([f(x, t) for x in X])


class dMOP2(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import dMOP2 as f
        out["F"] = np.array([f(x, t) for x in X])


class dMOP3(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import dMOP3 as f
        out["F"] = np.array([f(x, t) for x in X])


class HE2(DynamicCEC2015):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import HE2 as f
        out["F"] = np.array([f(x, t) for x in X])


class HE7(DynamicCEC2015):

    def __init__(self, n_var=10, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import HE7 as f
        out["F"] = np.array([f(x, t) for x in X])


class HE9(DynamicCEC2015):

    def __init__(self, n_var=10, **kwargs):
        super().__init__(n_var=n_var, **kwargs)

    def _evaluate(self, X, out, *args, **kwargs):
        t = self.time
        from pymoo.vendor.gta import HE9 as f
        out["F"] = np.array([f(x, t) for x in X])



from math import floor

import numpy as np
from pymoo.util.remote import Remote
from scipy.spatial.distance import cdist

from pydmoo.problems.dyn import DynamicTestProblem
from pydmoo.problems.dynamic.df import get_PF


def knee_point(solved_pf: np.ndarray):
    ideal_point = np.min(solved_pf, axis=0)
    distances = cdist(solved_pf, [ideal_point])
    knee_idx = np.argmax(distances)
    return solved_pf[knee_idx]


def G_t(t):
    # $G(t) = \\sin(0.5\\pi t)$
    return np.sin(0.5 * np.pi * t)


def H_t(t):
    # $H(t) = 1.5 + G(t)$
    return 1.5 + G_t(t)


def alpha_t(t):
    # $\\alpha_t = 5\\cos(0.5\\pi t)$
    return 5 * np.cos(0.5 * np.pi * t)


def beta_t(t):
    # $\\beta_t = 0.2 + 2.8|G(t)|$
    return 0.2 + 2.8 * np.abs(G_t(t))


def omega_t(t):
    # $\\omega_t = \\lfloor 10G(t) \\rfloor$
    return np.floor(10 * G_t(t))


def a_t(t):
    # $a(t) = \\sin(0.5\\pi t)$
    return np.sin(0.5 * np.pi * t)


def b_t(t):
    # $b(t) = 1 + |\\cos(0.5\\pi t)|$
    return 1 + np.abs(np.cos(0.5 * np.pi * t))


def y_t(x1, t):
    # $y_t(x_1) = 0.5 + G(t)(x_1 - 0.5)$
    return 0.5 + G_t(t) * (x1 - 0.5)


def p_t(t):
    # $p_t = \\lfloor 6G(t) \\rfloor$
    return np.floor(6 * G_t(t))


class GTS(DynamicTestProblem):

    def __init__(self,
                 part_idx,
                 bounds,
                 matrix_case="one",
                 add_time_perturbation=True,
                 n_var=10,
                 nt=10,
                 taut=20,
                 **kwargs):
        super().__init__(nt,
                         taut,
                         add_time_perturbation=add_time_perturbation,
                         n_var=n_var,
                         n_obj=2,
                         xl=0,
                         xu=1,
                         **kwargs)

        self.time_linkage = 1

        self.sub_vec_1, self.sub_vec_2, self.sub_vec_3 = self._partition_dimension(part_idx)

        # positive semidefinite matrices
        if matrix_case == "one":
            self.matrix_2 = np.eye(len(self.sub_vec_2))
            self.matrix_3 = np.eye(len(self.sub_vec_3))
        elif matrix_case == "two":
            self.matrix_2 = np.diag([i + 1 for i in range(len(self.sub_vec_2))])
            self.matrix_3 = np.diag([i + 1 for i in range(len(self.sub_vec_3))])
        elif matrix_case == "three":
            diag_matrix = np.diag(len(self.sub_vec_2) + np.arange(len(self.sub_vec_2)))
            ones_matrix = np.ones((len(self.sub_vec_2), len(self.sub_vec_2)))
            self.matrix_2 = np.where(np.eye(len(self.sub_vec_2), dtype=bool), diag_matrix, ones_matrix)

            diag_matrix = np.diag(len(self.sub_vec_3) + np.arange(len(self.sub_vec_3)))
            ones_matrix = np.ones((len(self.sub_vec_3), len(self.sub_vec_3)))
            self.matrix_3 = np.where(np.eye(len(self.sub_vec_3), dtype=bool), diag_matrix, ones_matrix)
        else:
            raise ValueError(f"{matrix_case} must be `one`, `two` or `three`.")

        # norm
        self.p = 1

        self.xl[self.sub_vec_1] = bounds[0][0]
        self.xu[self.sub_vec_1] = bounds[0][1]
        self.xl[self.sub_vec_2] = bounds[1][0]
        self.xu[self.sub_vec_2] = bounds[1][1]
        self.xl[self.sub_vec_3] = bounds[2][0]
        self.xu[self.sub_vec_3] = bounds[2][1]

    def _calc_pareto_front(self, *args, **kwargs):
        return Remote.get_instance().load("pydmoo", "pf", "GTS", f"{self.__class__.__name__}.pf")

    def _calc_pareto_set(self, *args, **kwargs):
        return Remote.get_instance().load("pydmoo", "ps", "GTS", f"{self.__class__.__name__}.ps")

    # Designed to handle time-linkage properties within the GTS test suites.
    def cal(self, F):
        self.time_linkage = 1 + np.linalg.norm(knee_point(self.pareto_front()) - knee_point(F))

    def _partition_dimension(self, part_idx):
        fd = floor(self.n_var / 2)
        sub_vec_1 = range(0, part_idx)
        sub_vec_2 = range(part_idx, fd + (part_idx - 1))
        sub_vec_3 = range(fd + (part_idx - 1), self.n_var)
        return sub_vec_1, sub_vec_2, sub_vec_3


class GTS1(GTS):
    r"""GTS1 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = x_1 \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)(1 -  (\frac{x_1}{g(\mathbf{x},t)})^{H(t)})
            \end{cases}
    \end{equation}

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = x_1 \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)(1 -  (\frac{x_1}{g(\mathbf{x},t)})^{H(t)})
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}}
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1)$, $\mathbf{x}_{II,1} = (x_2, \cdots, x_{\lfloor\frac{D}{2}\rfloor})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 1}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \cos(0.5\pi t)$ and $h_2(\mathbf{x}_I, t) = G(t) + x_1^{H(t)}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1] \times [-1,1]^{\lfloor\frac{D}{2}\rfloor -1} \times  [-1, 2]^{\lceil\frac{D}{2}\rceil}$.

    - Pareto set (PS)

    ![GTS1 PS](../../figs/PS/GTS1.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS1 PF](../../figs/PF/GTS1.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=1, bounds=((0, 1), (-1, 1), (-1, 2)), **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = np.cos(0.5 * np.pi * self.time)
        xj = G_t(self.time) + np.power(x[:, 0], H_t(self.time))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        f1 = x[:, 0]
        f2 = g * (1 - np.power(x[:, 0] / g, H_t(self.time)))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        x = np.linspace(0, 1, n_pareto_points)
        f1 = x
        f2 = 1 - np.power(x, H_t(self.time))
        return np.array([f1, f2]).T

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = np.full(len(x_vec1), np.cos(0.5 * np.pi * self.time))
        x_vec3 = G_t(self.time) + np.power(x_vec1, H_t(self.time))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


class GTS2(GTS):
    r"""GTS2 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = 0.5x_1+x_2 \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)(2.8 - (\frac{0.5x_1+x_2}{g(\mathbf{x},t)})^{H(t)})
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}}
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1, x_2)$, $\mathbf{x}_{II,1} = (x_3, \cdots, x_{\lfloor\frac{D}{2}\rfloor + 1})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 2}, \cdots, x_D)$, $c = \cot(3\pi t^2), \text{when~} t^2 \neq \frac{n}{3}, n \in \mathbb{Z}, c = 1e-32, \text{otherwise}$,
    $h_1(\mathbf{x}_I, t) = \frac{1}{\pi}\left\vert{\arctan(c)}\right\vert$ and $h_2(\mathbf{x}_I, t) = G(t) + x_1^{H(t)}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1]^2 \times [0,1]^{\lfloor\frac{D}{2}\rfloor -1} \times  [-1, 2]^{\lceil\frac{D}{2}\rceil-1}$.

    - Pareto set (PS)

    ![GTS2 PS](../../figs/PS/GTS2.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS2 PF](../../figs/PF/GTS2.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=2, bounds=((0, 1), (0, 1), (-1, 2)), **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        t = 3 * np.pi * self.time ** 2
        cot = np.cos(t) / (np.sin(t) + (np.sin(t) == 0) * 1e-32)

        xi = (1 / np.pi) * np.abs(np.arctan(cot))
        xj = G_t(self.time) + np.power(x[:, 0], H_t(self.time))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        f1 = 0.5 * x[:, 0] + x[:, 1]
        f2 = g * (2.8 - np.power(f1 / g, H_t(self.time)))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')

        f1 = 0.5 * x1 + x2
        f2 = 2.8 - np.power(f1, H_t(self.time))
        return get_PF(np.array([f1, f2]), False)

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        t = 3 * np.pi * self.time ** 2
        cot = np.cos(t) / (np.sin(t) + (np.sin(t) == 0) * 1e-32)

        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = np.full(len(x_vec1), (1 / np.pi) * np.abs(np.arctan(cot)))
        x_vec3 = G_t(self.time) + np.power(x_vec1, H_t(self.time))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


class GTS3(GTS):
    r"""GTS3 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = g(x)(x_1 +  0.1\sin(3\pi x_1))^{\beta_t} \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)(1 - x_1 + 0.1\sin(3\pi  x_1))^{\beta_t}
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}}
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1)$, $\mathbf{x}_{II,1} = (x_2, \cdots, x_{\lfloor\frac{D}{2}\rfloor})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 1}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \frac{G(t)\sin(4\pi x_1)}{1 + \left\vert{G(t)}\right\vert}$ and $h_2(\mathbf{x}_I, t) = G(t) + x_1^{H(t)}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1] \times [-1,1]^{\lfloor\frac{D}{2}\rfloor - 1} \times  [-1, 2]^{\lceil\frac{D}{2}\rceil}$.

    - Pareto set (PS)

    ![GTS3 PS](../../figs/PS/GTS3.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS3 PF](../../figs/PF/GTS3.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=1, bounds=((0, 1), (-1, 1), (-1, 2)), **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = (G_t(self.time) * np.sin(4 * np.pi * x[:, 0])) / (1 + np.abs(G_t(self.time)))
        xj = G_t(self.time) + np.power(x[:, 0], H_t(self.time))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        f1 = g * np.power(x[:, 0] + 0.1 * np.sin(3 * np.pi * x[:, 0]), beta_t(self.time))
        f2 = g * np.power(1 - x[:, 0] + 0.1 * np.sin(3 * np.pi * x[:, 0]), beta_t(self.time))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        x = np.linspace(0, 1, n_pareto_points)
        f1 = np.power(x + 0.1 * np.sin(3 * np.pi * x), beta_t(self.time))
        f2 = np.power(1 - x + 0.1 * np.sin(3 * np.pi * x), beta_t(self.time))

        return get_PF(np.array([f1, f2]), True)

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = (G_t(self.time) * np.sin(4 * np.pi * x_vec1)) / (1 + np.abs(G_t(self.time)))
        x_vec3 = G_t(self.time) + np.power(x_vec1, H_t(self.time))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


class GTS4(GTS):
    r"""GTS4 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = g(\mathbf{x},t)\frac{1 +  t}{x_1 + 3} \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)\frac{x_1 + 3}{1 + t}
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & - 0.5 + 0.25\sin(0.3\pi t)
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1)$, $\mathbf{x}_{II,1} = (x_2, \cdots, x_{\lfloor\frac{D}{2}\rfloor})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 1}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \left\vert{G(t)}\right\vert$ and $h_2(\mathbf{x}_I, t) = \frac{G(t)\sin(4\pi x_1)}{1 + \left\vert{G(t)}\right\vert}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1] \times [0,1]^{\lfloor\frac{D}{2}\rfloor -1} \times  [-1, 1]^{\lceil\frac{D}{2}\rceil}$.

    - Pareto set (PS)

    ![GTS4 PS](../../figs/PS/GTS4.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS4 PF](../../figs/PF/GTS4.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=1, bounds=((0, 1), (0, 1), (-1, 1)), **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = np.abs(G_t(self.time))
        xj = (G_t(self.time) * np.sin(4 * np.pi * x[:, 0])) / (1 + np.abs(G_t(self.time)))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        g += (-0.5 + 0.25 * np.sin(0.3 * np.pi * self.time))

        f1 = g * ((1 + self.time) / (x[:, 0] + 3))
        f2 = g * ((x[:, 0] + 3) / (1 + self.time))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        x = np.linspace(0, 1, n_pareto_points)
        g = 1 + (-0.5 + 0.25 * np.sin(0.3 * np.pi * self.time))

        f1 = g * ((1 + self.time) / (x + 3))
        f2 = g * ((x + 3) / (1 + self.time))

        return np.array([f1, f2]).T

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = np.full(len(x_vec1), np.abs(G_t(self.time)))
        x_vec3 = (G_t(self.time) * np.sin(4 * np.pi * x_vec1)) / (1 + np.abs(G_t(self.time)))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


class GTS5(GTS):
    r"""GTS5 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) =  g(\mathbf{x},t)((0.5x_1+x_2)) + 0.02\sin(\omega_t\pi (0.5x_1+x_2))) \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)(1.6 - (0.5x_1+x_2) + 0.02\sin(\omega_t\pi  (0.5x_1+x_2)))
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + 0.5 + 0.5G(t)
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1, x_2)$, $\mathbf{x}_{II,1} = (x_3, \cdots, x_{\lfloor\frac{D}{2}\rfloor + 1})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 2}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \cos(0.5\pi t)$ and $h_2(\mathbf{x}_I, t) = G(t) + x_1^{H(t)}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1]^2 \times [-1,1]^{\lfloor\frac{D}{2}\rfloor - 1} \times  [-1, 2]^{\lceil\frac{D}{2}\rceil-1}$.

    - Pareto set (PS)

    ![GTS5 PS](../../figs/PS/GTS5.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS5 PF](../../figs/PF/GTS5.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=2, bounds=((0, 1), (-1, 1), (-1, 2)), **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = np.cos(0.5 * np.pi * self.time)
        xj = G_t(self.time) + np.power(x[:, 0], H_t(self.time))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        g += 0.5 + 0.5 * G_t(self.time)

        x12 = 0.5 * x[:, 0] + x[:, 1]
        _sin = 0.02 * np.sin(omega_t(self.time) * np.pi * x12)
        f1 = g * (x12 + _sin)
        f2 = g * (1.6 - x12 + _sin)

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        g = 1 + 0.5 + 0.5 * G_t(self.time)

        x12 = 0.5 * x1 + x2
        _sin = 0.02 * np.sin(omega_t(self.time) * np.pi * x12)
        f1 = g * (x12 + _sin)
        f2 = g * (1.6 - x12 + _sin)

        return get_PF(np.array([f1, f2]), True)

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = np.full(len(x_vec1), np.cos(0.5 * np.pi * self.time))
        x_vec3 = G_t(self.time) + np.power(x_vec1, H_t(self.time))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


class GTS6(GTS):
    r"""GTS6 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = x_1 \\
            f_2(\mathbf{x},t) =  g(\mathbf{x},t)(1 - (\frac{x_1}{g(\mathbf{x},t)})^{H(t)})
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}}
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1)$, $\mathbf{x}_{II,1} = (x_2, \cdots, x_{\lfloor\frac{D}{2}\rfloor})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 1}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \cos(0.5\pi t)$ and $h_2(\mathbf{x}_I, t) = G(t) + x_1^{H(t)}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1] \times [-1,1]^{\lfloor\frac{D}{2}\rfloor -1} \times  [-1, 2]^{\lceil\frac{D}{2}\rceil}$.

    - Pareto set (PS)

    ![GTS6 PS](../../figs/PS/GTS6.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS6 PF](../../figs/PF/GTS6.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=1, bounds=((0, 1), (-1, 1), (-1, 2)), **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = self.time_linkage * np.cos(0.5 * np.pi * self.time)
        xj = self.time_linkage * (G_t(self.time) + np.power(x[:, 0], H_t(self.time)))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        f1 = x[:, 0]
        f2 = g * (1 - np.power(f1 / g, H_t(self.time)))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        f1 = np.linspace(0, 1, n_pareto_points)
        f2 = 1 - np.power(f1, H_t(self.time))
        return np.array([f1, f2]).T

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = np.full(len(x_vec1), np.cos(0.5 * np.pi * self.time))
        x_vec3 = G_t(self.time) + np.power(x_vec1, H_t(self.time))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


class GTS7(GTS):
    r"""GTS7 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) =  g(\mathbf{x},t)\left\vert{x_1-a_t}\right\vert^{H(t)} \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)\left\vert{x_1-a_t-b_t}\right\vert^{H(t)}
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}}
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1)$, $\mathbf{x}_{II,1} = (x_2, \cdots, x_{\lfloor\frac{D}{2}\rfloor})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 1}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \cos(0.5\pi t)$ and $h_2(\mathbf{x}_I, t) = \frac{1}{1 + e^{\alpha_t(x_1 - 0.5)}}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[-1,2.5] \times [-1,1]^{\lfloor\frac{D}{2}\rfloor - 1} \times  [0, 1]^{\lceil\frac{D}{2}\rceil}$.

    - Pareto set (PS)

    ![GTS7 PS](../../figs/PS/GTS7.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS7 PF](../../figs/PF/GTS7.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=1, bounds=((-1, 2.5), (-1, 1), (0, 1)), **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = self.time_linkage * np.cos(0.5 * np.pi * self.time)
        xj = self.time_linkage * 1 / (1 + np.exp(alpha_t(self.time) * (x[:, 0] - 0.5)))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        f1 = g * np.power(np.abs(x[:, 0] - a_t(self.time)), H_t(self.time))
        f2 = g * np.power(np.abs(x[:, 0] - a_t(self.time) - b_t(self.time)), H_t(self.time))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        x = np.linspace(a_t(self.time), a_t(self.time) + b_t(self.time), n_pareto_points)

        f1 = np.power(np.abs(x - a_t(self.time)), H_t(self.time))
        f2 = np.power(np.abs(x - a_t(self.time) - b_t(self.time)), H_t(self.time))

        return np.array([f1, f2]).T

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(a_t(self.time), a_t(self.time) + b_t(self.time), n_pareto_points)
        x_vec2 = np.full(len(x_vec1), np.cos(0.5 * np.pi * self.time))
        x_vec3 = 1 / (1 + np.exp(alpha_t(self.time) * (x_vec1 - 0.5)))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


class GTS8(GTS):
    r"""GTS8 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = (0.5x_1+x_2)                                                         \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)(2.8 - (\frac{(0.5x_1+x_2)}{g(\mathbf{x},t)})^{H(t)}) \\
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + 0.25\left\vert{\cos(0.3 \pi t)}\right\vert
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1, x_2)$, $\mathbf{x}_{II,1} = (x_3, \cdots, x_{\lfloor\frac{D}{2}\rfloor + 1})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 2}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \frac{1}{1 + e^{\alpha _t(x_1 - 0.5)}}$ and $h_2(\mathbf{x}_I, t) = G(t) + x_1^{H(t)}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1]^2 \times [0,1]^{\lfloor\frac{D}{2}\rfloor -1} \times  [-1, 2]^{\lceil\frac{D}{2}\rceil-1}$.

    - Pareto set (PS)

    ![GTS8 PS](../../figs/PS/GTS8.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS8 PF](../../figs/PF/GTS8.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=2, bounds=((0, 1), (0, 1), (-1, 2)), **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = self.time_linkage * 1 / (1 + np.exp(alpha_t(self.time) * (x[:, 0] - 0.5)))
        xj = self.time_linkage * (G_t(self.time) + np.power(x[:, 0], H_t(self.time)))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        g += 0.25 * np.abs(np.cos(0.3 * np.pi * self.time))

        f1 = 0.5 * x[:, 0] + x[:, 1]
        f2 = g * (2.8 - np.power(f1 / g, H_t(self.time)))

        out["F"] = np.column_stack([f1, f2])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')

        f1 = 0.5 * x1 + x2
        g = 1 + 0.25 * np.abs(np.cos(0.3 * np.pi * self.time))
        f2 = g * (2.8 - np.power(f1 / g, H_t(self.time)))
        return get_PF(np.array([f1, f2]), False)

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = 1 / (1 + np.exp(alpha_t(self.time) * (x_vec1 - 0.5)))
        x_vec3 = G_t(self.time) + np.power(x_vec1, H_t(self.time))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


# modified DF12
class GTS9(GTS):
    r"""GTS9 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = g(\mathbf{x},t)\cos(0.5\pi  x_1)\cos(0.5\pi x_2) \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)\cos(0.5\pi x_1)\sin(0.5\pi x_2)  \\
            f_3(\mathbf{x},t) = g(\mathbf{x},t)\sin(0.5\pi x_1)
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \left\vert{\cos(0.27\pi t)}\right\vert
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1, x_2)$, $\mathbf{x}_{II,1} = (x_3, \cdots, x_{\lfloor\frac{D}{2}\rfloor + 1})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 2}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \frac{1}{1+e^{\alpha_t(x_1 - 0.5)}}$ and $h_2(\mathbf{x}_I, t) = \sin(tx_1)$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1]^2 \times [0,1]^{\lfloor\frac{D}{2}\rfloor - 1} \times  [-1, 1]^{\lceil\frac{D}{2}\rceil - 1}$.

    - Pareto set (PS)

    ![GTS9 PS](../../figs/PS/GTS9.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS9 PF](../../figs/PF/GTS9.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=2, bounds=((0, 1), (0, 1), (-1, 1)), **kwargs)
        self.n_obj = 3

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = 1 / (1 + np.exp(alpha_t(self.time) * (x[:, 0] - 0.5)))
        xj = np.sin(self.time * x[:, 0])

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        g += np.abs(np.cos(0.27 * np.pi * self.time))

        f1 = g * np.cos(0.5 * np.pi * x[:, 1]) * np.cos(0.5 * np.pi * x[:, 0])
        f2 = g * np.sin(0.5 * np.pi * x[:, 1]) * np.cos(0.5 * np.pi * x[:, 0])
        f3 = g * np.sin(0.5 * np.pi * x[:, 0])

        out["F"] = np.column_stack([f1, f2, f3])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')

        g = 1
        g += np.abs(np.cos(0.27 * np.pi * self.time))

        f1 = np.multiply(np.multiply(g, np.cos(0.5 * np.pi * x2)), np.cos(0.5 * np.pi * x1))
        f2 = np.multiply(np.multiply(g, np.sin(0.5 * np.pi * x2)), np.cos(0.5 * np.pi * x1))
        f3 = np.multiply(g, np.sin(0.5 * np.pi * x1))

        return get_PF(np.array([f1, f2, f3]), True)

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = 1 / (1 + np.exp(alpha_t(self.time) * (x_vec1 - 0.5)))
        x_vec3 = np.sin(self.time * x_vec1)
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


# modified DF13
class GTS10(GTS):
    r"""GTS10 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) =  g(\mathbf{x},t)\cos^2(0.5\pi x_1) \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)\cos^2(0.5\pi x_2)  \\
            f_3(\mathbf{x},t) = g(\mathbf{x},t)\sum_{j = 1}^{2}(\sin^2(0.5\pi x_j) + \sin(0.5\pi  x_j)\cos^2(\lfloor6G(t)\rfloor \pi x_j))
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}}
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1, x_2)$, $\mathbf{x}_{II,1} = (x_3, \cdots, x_{\lfloor\frac{D}{2}\rfloor + 1})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 2}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \left\vert{G(t)}\right\vert$ and $h_2(\mathbf{x}_I, t) = -0.5 + \frac{\left\vert{G(t)\sin(4\pi x_1)}\right\vert}{0.5(1+\left\vert{G(t)}\right\vert)}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,%
    the search space is $[0,1]^2 \times [0,1]^{\lfloor\frac{D}{2}\rfloor - 1} \times  [-1, 1]^{\lceil\frac{D}{2}\rceil - 1}$.

    - Pareto set (PS)

    ![GTS10 PS](../../figs/PS/GTS10.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS10 PF](../../figs/PF/GTS10.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=2, bounds=((0, 1), (0, 1), (-1, 1)), **kwargs)
        self.n_obj = 3

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = np.abs(G_t(self.time))
        xj = -0.5 + np.abs(G_t(self.time) * np.sin(4 * np.pi * x[:, 1])) / (0.5 * (1 + np.abs(G_t(self.time))))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        f1 = g * np.cos(0.5 * np.pi * x[:, 0]) ** 2
        f2 = g * np.cos(0.5 * np.pi * x[:, 1]) ** 2

        _sin1 = np.sin(0.5 * np.pi * x[:, 0])
        _sin2 = np.sin(0.5 * np.pi * x[:, 1])
        _cos1 = np.cos(p_t(self.time) * np.pi * x[:, 0])
        _cos2 = np.cos(p_t(self.time) * np.pi * x[:, 1])
        f3 = g * (_sin1 ** 2 + _sin1 * _cos1 ** 2 + _sin2 ** 2 + _sin2 * _cos2 ** 2)

        out["F"] = np.column_stack([f1, f2, f3])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')
        G = np.sin(0.5 * np.pi * self.time)
        p = np.floor(6 * G)

        f1 = np.cos(0.5 * np.pi * x1) ** 2
        f2 = np.cos(0.5 * np.pi * x2) ** 2
        f3 = np.sin(0.5 * np.pi * x1) ** 2 + np.sin(0.5 * np.pi * x1) * np.cos(p * np.pi * x1) ** 2 + np.sin(
            0.5 * np.pi * x2) ** 2 + np.sin(0.5 * np.pi * x2) * np.cos(p * np.pi * x2) ** 2

        return get_PF(np.array([f1, f2, f3]), True)

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = np.full(len(x_vec1), np.abs(G_t(self.time)))
        x_vec3 = -0.5 + np.abs(G_t(self.time) * np.sin(4 * np.pi * x_vec1)) / (0.5 * (1 + np.abs(G_t(self.time))))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


# modified DF14
class GTS11(GTS):
    r"""GTS11 test problem.

    - Inherits all parameters from parent class GTS.

    Attributes
    ----------
    name : str
        Problem name, default is 'GTS1'
    n_var : int
        Number of decision variables
    n_obj : int
        Number of objective functions
    time_linkage : bool
        Whether the problem has time linkage

    Notes
    -----
    - Mathematical Formulation:

    \begin{equation}
        \text{min}
        \begin{cases}
            f_1(\mathbf{x},t) = g(\mathbf{x},t)(1.05 - y +  0.05\sin(6\pi y))                           \\
            f_2(\mathbf{x},t) = g(\mathbf{x},t)(1.05 - x_2 + 0.05\sin(6\pi x_2))(y +  0.05\sin(6\pi y)) \\
            f_3(\mathbf{x},t) = g(\mathbf{x},t)(x_2 + 0.05\sin(6\pi x_2))(y + 0.05\sin(6\pi y))
        \end{cases}
    \end{equation}

    with

    \begin{equation*}
        \begin{split}
            g(\mathbf{x},t) = 1
             & + \Bigl(\bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,1}(t) \bigl(\mathbf{x}_{II,1} - h_1(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}} \\
             & + \Bigl(\bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)^T \mathbf{R}_{II,2}(t) \bigl(\mathbf{x}_{II,2} - h_2(\mathbf{x}_I)\bigr)\Bigr)^{\frac{1}{p}}
        \end{split}
    \end{equation*}

    where $p \geq 1$, $\mathbf{x}_I = (x_1, x_2)$, $\mathbf{x}_{II,1} = (x_3, \cdots, x_{\lfloor\frac{D}{2}\rfloor + 1})$ and $\mathbf{x}_{II,2} = (x_{\lfloor\frac{D}{2}\rfloor + 2}, \cdots, x_D)$,
    $h_1(\mathbf{x}_I, t) = \left\vert{G(t)}\right\vert$ and $h_2(\mathbf{x}_I, t) = G(t) + x_1^{H(t)}$,
    $\mathbf{R}_{II,1}(t)$ and $\mathbf{R}_{II,2}(t)$ are symmetric positive semidefinite matrices in the $t$-th environment,
    the search space is $[0,1]^2 \times [0,1]^{\lfloor\frac{D}{2}\rfloor - 1} \times  [-1, 2]^{\lceil\frac{D}{2}\rceil - 1}$.

    - Pareto set (PS)

    ![GTS11 PS](../../figs/PS/GTS11.png){: width="400px" height="300px"}

    - Pareto front (PF)

    ![GTS11 PF](../../figs/PF/GTS11.png){: width="400px" height="300px"}
    """
    def __init__(self, **kwargs):
        super().__init__(part_idx=2, bounds=((0, 1), (0, 1), (-1, 2)), **kwargs)
        self.n_obj = 3

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate."""
        xi = np.abs(G_t(self.time))
        xj = G_t(self.time) + np.power(x[:, 0], H_t(self.time))

        x2, x3 = (x[:, self.sub_vec_2] - xi.reshape(-1, 1)), (x[:, self.sub_vec_3] - xj.reshape(-1, 1))

        # quadratic form
        diag_x2_R2_x2T = np.einsum('ij,jk,ik->i', x2, self.matrix_2, x2)
        diag_x3_R3_x3T = np.einsum('ij,jk,ik->i', x3, self.matrix_3, x3)
        g = 1 + np.power(diag_x2_R2_x2T, self.p) + np.power(diag_x3_R3_x3T, self.p)

        y = y_t(x[:, 0], self.time)
        f1 = g * (1.05 - y + 0.05 * np.sin(6 * np.pi * y))
        f2 = g * (1.05 - x[:, 1] + 0.05 * np.sin(6 * np.pi * x[:, 1])) * (y + 0.05 * np.sin(6 * np.pi * y))
        f3 = g * (x[:, 1] + 0.05 * np.sin(6 * np.pi * x[:, 1])) * (y + 0.05 * np.sin(6 * np.pi * y))

        out["F"] = np.column_stack([f1, f2, f3])

    def _calc_pareto_front(self, *args, n_pareto_points=100, **kwargs):
        """Pareto front."""
        H = 20
        x1, x2 = np.meshgrid(np.linspace(0, 1, H), np.linspace(0, 1, H), indexing='xy')

        y = y_t(x1, self.time)
        f1 = 1.05 - y + 0.05 * np.sin(6 * np.pi * y)
        f2 = np.multiply(1.05 - x2 + 0.05 * np.sin(6 * np.pi * x2), y + 0.05 * np.sin(6 * np.pi * y))
        f3 = np.multiply(x2 + 0.05 * np.sin(6 * np.pi * x2), y + 0.05 * np.sin(6 * np.pi * y))

        return get_PF(np.array([f1, f2, f3]), False)

    def _calc_pareto_set(self, *args, n_pareto_points=100, **kwargs):
        """Pareto set."""
        x_vec1 = np.linspace(0, 1, n_pareto_points)
        x_vec2 = np.full(len(x_vec1), np.abs(G_t(self.time)))
        x_vec3 = G_t(self.time) + np.power(x_vec1, H_t(self.time))
        X, Y, Z = x_vec1, x_vec2, x_vec3
        return np.array([X.flatten(order='F'), Y.flatten(order='F'), Z.flatten(order='F')]).T


class GTS1_2(GTS1):
    """GTS1_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS2_2(GTS1):
    """GTS2_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS3_2(GTS1):
    """GTS3_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS4_2(GTS1):
    """GTS4_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS5_2(GTS1):
    """GTS5_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS6_2(GTS1):
    """GTS6_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS7_2(GTS1):
    """GTS7_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS8_2(GTS1):
    """GTS8_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS9_2(GTS1):
    """GTS9_2 test problem."""
    def __init__(self, matix_case="two", **kwargs):
        super().__init__(matrix_case=matix_case, **kwargs)


class GTS10_2(GTS1):
    """GTS10_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS11_2(GTS1):
    """GTS11_2 test problem."""
    def __init__(self, matrix_case="two", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS1_3(GTS1):
    """GTS1_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS2_3(GTS1):
    """GTS2_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS3_3(GTS1):
    """GTS3_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS4_3(GTS1):
    """GTS4_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS5_3(GTS1):
    """GTS5_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS6_3(GTS1):
    """GTS6_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS7_3(GTS1):
    """GTS7_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS8_3(GTS1):
    """GTS8_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS9_3(GTS1):
    """GTS9_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS10_3(GTS1):
    """GTS10_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)


class GTS11_3(GTS1):
    """GTS11_3 test problem."""
    def __init__(self, matrix_case="three", **kwargs):
        super().__init__(matrix_case=matrix_case, **kwargs)



import numpy as np


class ARModel:
    """
    Autoregressive (AR) model implementation from scratch.

    Parameters
    ----------
    p : int
        Order of the AR model (number of lagged observations)
    """

    def __init__(self, p):
        self.p = p  # AR order
        self.coef_ = None  # Model coefficients (including intercept if present)
        self.resid_ = None  # Residuals after fitting

    def fit(self, y, trend="c"):
        """
        Fit AR model to time series data.

        Parameters
        ----------
        y : ndarray, shape (M,)
            Time series data of length M
        trend : str, optional
            'c' for constant (default), 'n' for no intercept

        Returns
        -------
        self : returns an instance of self
        """
        M = len(y)
        if M <= self.p:
            raise ValueError(f"Time series length M={M} must be > order p={self.p}")

        # Construct design matrix X and target vector Y
        X = np.zeros((M - self.p, self.p))
        for i in range(self.p):
            X[:, i] = y[(self.p - i - 1): (M - i - 1)]

        Y = y[self.p:]

        # Add intercept if specified
        if trend == "c":
            X = np.column_stack([np.ones(X.shape[0]), X])

        # Solve least squares problem
        self.coef_ = np.linalg.lstsq(X, Y, rcond=None)[0]

        # Store residuals
        self.resid_ = Y - X @ self.coef_

        return self

    def predict(self, y, steps=1):
        """
        Predict future values using the fitted AR model.

        Parameters
        ----------
        y : ndarray, shape (K,)
            Input sequence (K >= p)
        steps : int, optional
            Number of steps to predict (default=1)

        Returns
        -------
        predictions : ndarray, shape (steps,)
            Predicted values
        """
        if len(y) < self.p:
            raise ValueError(f"Input length {len(y)} must be >= model order {self.p}")
        if self.coef_ is None:
            raise ValueError("Model must be fitted before prediction")

        predictions = np.zeros(steps)
        history = y[-self.p:].copy()  # Last p observations

        for i in range(steps):
            # Prepare input vector
            x = history[-self.p:][::-1]  # Latest p observations in reverse order

            # Add intercept if present (coef[0] is intercept)
            if len(self.coef_) == self.p + 1:
                x = np.insert(x, 0, 1)

            # Make prediction
            pred = np.dot(x, self.coef_)
            predictions[i] = pred

            # Update history
            history = np.append(history, pred)

        return predictions



from pydmoo.algorithms.modern.moeadde_imkt_clstm import (
    MOEADDEIMicLSTM1003,
    MOEADDEIMicLSTM1005,
    MOEADDEIMicLSTM1007,
    MOEADDEIMicLSTM1009,
    MOEADDEIMicLSTM1503,
    MOEADDEIMicLSTM1505,
    MOEADDEIMicLSTM1507,
    MOEADDEIMicLSTM1509,
    MOEADDEIMicLSTM1511,
    MOEADDEIMicLSTM1513,
)
from pydmoo.algorithms.modern.moeadde_imkt_lstm import (
    MOEADDEIMiLSTM1003,
    MOEADDEIMiLSTM1005,
    MOEADDEIMiLSTM1007,
    MOEADDEIMiLSTM1009,
    MOEADDEIMiLSTM1503,
    MOEADDEIMiLSTM1505,
    MOEADDEIMiLSTM1507,
    MOEADDEIMiLSTM1509,
    MOEADDEIMiLSTM1511,
    MOEADDEIMiLSTM1513,
)
from pydmoo.algorithms.modern.nsga2_imkt_clstm import (
    NSGA2IMicLSTM1003,
    NSGA2IMicLSTM1005,
    NSGA2IMicLSTM1007,
    NSGA2IMicLSTM1009,
    NSGA2IMicLSTM1503,
    NSGA2IMicLSTM1505,
    NSGA2IMicLSTM1507,
    NSGA2IMicLSTM1509,
    NSGA2IMicLSTM1511,
    NSGA2IMicLSTM1513,
)
from pydmoo.algorithms.modern.nsga2_imkt_lstm import (
    NSGA2IMiLSTM1003,
    NSGA2IMiLSTM1005,
    NSGA2IMiLSTM1007,
    NSGA2IMiLSTM1009,
    NSGA2IMiLSTM1503,
    NSGA2IMiLSTM1505,
    NSGA2IMiLSTM1507,
    NSGA2IMiLSTM1509,
    NSGA2IMiLSTM1511,
    NSGA2IMiLSTM1513,
)

ALGORITHMS_ABLATION = {
    "MOEADDEIMicLSTM1003": MOEADDEIMicLSTM1003,
    "MOEADDEIMicLSTM1005": MOEADDEIMicLSTM1005,
    "MOEADDEIMicLSTM1007": MOEADDEIMicLSTM1007,
    "MOEADDEIMicLSTM1009": MOEADDEIMicLSTM1009,
    "MOEADDEIMicLSTM1503": MOEADDEIMicLSTM1503,
    "MOEADDEIMicLSTM1505": MOEADDEIMicLSTM1505,
    "MOEADDEIMicLSTM1507": MOEADDEIMicLSTM1507,
    "MOEADDEIMicLSTM1509": MOEADDEIMicLSTM1509,
    "MOEADDEIMicLSTM1511": MOEADDEIMicLSTM1511,
    "MOEADDEIMicLSTM1513": MOEADDEIMicLSTM1513,

    "MOEADDEIMiLSTM1003": MOEADDEIMiLSTM1003,
    "MOEADDEIMiLSTM1005": MOEADDEIMiLSTM1005,
    "MOEADDEIMiLSTM1007": MOEADDEIMiLSTM1007,
    "MOEADDEIMiLSTM1009": MOEADDEIMiLSTM1009,
    "MOEADDEIMiLSTM1503": MOEADDEIMiLSTM1503,
    "MOEADDEIMiLSTM1505": MOEADDEIMiLSTM1505,
    "MOEADDEIMiLSTM1507": MOEADDEIMiLSTM1507,
    "MOEADDEIMiLSTM1509": MOEADDEIMiLSTM1509,
    "MOEADDEIMiLSTM1511": MOEADDEIMiLSTM1511,
    "MOEADDEIMiLSTM1513": MOEADDEIMiLSTM1513,

    "NSGA2IMicLSTM1003": NSGA2IMicLSTM1003,
    "NSGA2IMicLSTM1005": NSGA2IMicLSTM1005,
    "NSGA2IMicLSTM1007": NSGA2IMicLSTM1007,
    "NSGA2IMicLSTM1009": NSGA2IMicLSTM1009,
    "NSGA2IMicLSTM1503": NSGA2IMicLSTM1503,
    "NSGA2IMicLSTM1505": NSGA2IMicLSTM1505,
    "NSGA2IMicLSTM1507": NSGA2IMicLSTM1507,
    "NSGA2IMicLSTM1509": NSGA2IMicLSTM1509,
    "NSGA2IMicLSTM1511": NSGA2IMicLSTM1511,
    "NSGA2IMicLSTM1513": NSGA2IMicLSTM1513,

    "NSGA2IMiLSTM1003": NSGA2IMiLSTM1003,
    "NSGA2IMiLSTM1005": NSGA2IMiLSTM1005,
    "NSGA2IMiLSTM1007": NSGA2IMiLSTM1007,
    "NSGA2IMiLSTM1009": NSGA2IMiLSTM1009,
    "NSGA2IMiLSTM1503": NSGA2IMiLSTM1503,
    "NSGA2IMiLSTM1505": NSGA2IMiLSTM1505,
    "NSGA2IMiLSTM1507": NSGA2IMiLSTM1507,
    "NSGA2IMiLSTM1509": NSGA2IMiLSTM1509,
    "NSGA2IMiLSTM1511": NSGA2IMiLSTM1511,
    "NSGA2IMiLSTM1513": NSGA2IMiLSTM1513,
}



import os
import random

import numpy as np
import torch


class TimeSeriesBase:
    """Base class for time series forecasting models.

    This class provides common functionality for time series data processing,
    including data conversion, training data preparation, and prediction methods.

    Attributes
    ----------
    sequence_length : int
        Number of historical time steps used for each prediction.
    device : torch.device
        Computation device (CPU or GPU) for model inference.
    model_type : str
        Type of model architecture ('lstm' or 'transformer').
    """

    def __init__(self, sequence_length: int, device: torch.device, model_type: str = "lstm") -> None:
        """Initialize the time series base class.

        Parameters
        ----------
        sequence_length : int
            Number of historical time steps used for each prediction.
        device : torch.device
            Computation device (CPU or GPU) for model inference.
        model_type : str, optional
            Type of model architecture, by default "lstm".
            Supported values: 'lstm', 'transformer'

        Raises
        ------
        ValueError
            If model_type is not supported.
        """
        if model_type not in ["lstm", "transformer"]:
            raise ValueError(f"Unsupported model_type: {model_type}. Use 'lstm' or 'transformer'.")

        self.sequence_length = sequence_length
        self.device = device
        self.model_type = model_type

    def _set_random_seed(self, seed, deterministic=True):
        os.environ['PYTHONHASHSEED'] = str(seed)

        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)

        if self.device == "cpu":
            os.environ['CUDA_VISIBLE_DEVICES'] = ''
            torch.set_default_device('cpu')
            torch.set_default_dtype(torch.float32)

        elif torch.cuda.is_available():
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            if deterministic:
                torch.backends.cudnn.deterministic = True
                torch.backends.cudnn.benchmark = False
                if hasattr(torch, 'use_deterministic_algorithms'):
                    torch.use_deterministic_algorithms(True)

    def set_seed(self, seed: int) -> None:
        """Set random seed after initialization.

        Parameters
        ----------
        seed : int
            Random seed value.
        """
        self.seed = seed
        self._set_random_seed(seed)

    def convert_to_tensor(self, time_series_data: list[list[float]]) -> torch.Tensor:
        """Convert input time series data to PyTorch tensor.

        Parameters
        ----------
        time_series_data : list[list[float]]
            Input time series as list of feature vectors.
            Shape: (n_timesteps, n_features)

        Returns
        -------
        series_data : torch.Tensor
            Converted tensor of shape (n_timesteps, n_features)

        Raises
        ------
        ValueError
            If input data is invalid, not 2D, or insufficient for sequence length.

        Notes
        -----
        Performs comprehensive validation including:
        - Data type checking
        - Array dimensionality validation
        - Sequence length sufficiency check
        """
        # Input validation
        if not time_series_data or not all(isinstance(x, (int, float)) for row in time_series_data for x in row):
            raise ValueError("Invalid time series data")

        # Convert to numpy array first for efficient processing
        np_array = np.array(time_series_data, dtype=np.float32)

        # Validate array shape and dimensions
        if np_array.ndim != 2:
            raise ValueError(f"Expected 2D array, got {np_array.ndim}D array")

        n_timesteps, _ = np_array.shape  # (n_timesteps, n_features)

        # Validate sequence length requirement
        if self.sequence_length >= n_timesteps:
            raise ValueError(
                f"Sequence length {self.sequence_length} must be less than number of timesteps {n_timesteps}"
            )

        # Convert to PyTorch tensor
        series_data = torch.FloatTensor(np_array)
        return series_data

    def prepare_training_data(self, series_data: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """Prepare training sequences and targets using sliding window approach.

        Parameters
        ----------
        series_data : torch.Tensor
            Input time series tensor of shape (n_timesteps, n_features)

        Returns
        -------
        sequences_tensor : torch.Tensor
            Training sequences of shape (n_samples, sequence_length, n_features)
        targets_tensor : torch.Tensor
            Target values of shape (n_samples, n_features)

        Notes
        -----
        Uses sliding window method to create input-target pairs.
        Number of training samples = n_timesteps - sequence_length.

        The sliding window approach ensures temporal continuity in training data
        by creating overlapping sequences from the original time series.
        """
        n_timesteps, _ = series_data.shape

        # Calculate number of training samples
        n_samples = n_timesteps - self.sequence_length

        # Create sequence-target pairs using sliding window
        indices = torch.arange(n_samples).unsqueeze(1) + torch.arange(self.sequence_length)
        sequences_tensor = series_data[indices]  # shape: (n_samples, sequence_length, n_features)
        targets_tensor = series_data[self.sequence_length:]  # shape: (n_samples, n_features)

        return sequences_tensor, targets_tensor

    @staticmethod
    def create_training_dataloader(
        sequences_tensor: torch.Tensor, targets_tensor: torch.Tensor, batch_size: int = 32
    ) -> torch.utils.data.DataLoader:
        """Create DataLoader for time series training data.

        Parameters
        ----------
        sequences_tensor : torch.Tensor
            Input sequences of shape (n_samples, sequence_length, n_features)
        targets_tensor : torch.Tensor
            Target values of shape (n_samples, n_features)
        batch_size : int, optional
            Number of samples per batch, by default 32

        Returns
        -------
        dataloader : torch.utils.data.DataLoader
            Configured DataLoader for training

        Notes
        -----
        Key configurations:
        - shuffle=False: Maintains temporal order for time series data
        - pin_memory=False: Optimizes GPU data transfer
        - drop_last=False: Uses all available samples
        - num_workers=0: Avoids multiprocessing issues

        Temporal ordering is critical for time series data to preserve
        the sequential dependencies between data points.
        """
        dataset = torch.utils.data.TensorDataset(sequences_tensor, targets_tensor)

        dataloader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=False,  # Critical for time series data
            num_workers=0,
            pin_memory=False,
            drop_last=False,
        )

        return dataloader

    def predict_future(
        self, model, historical_data: torch.Tensor | list[list[float]], n_steps: int = 1
    ) -> torch.Tensor:
        """Generate multiple future predictions using iterative forecasting.

        Parameters
        ----------
        model : torch.nn.Module
            Trained model for time series prediction.
        historical_data : torch.Tensor | list[list[float]]
            Historical time series data of shape (n_timesteps, n_features).
        n_steps : int, optional
            Number of future time steps to predict, by default 1.

        Returns
        -------
        predictions : torch.Tensor
            Predicted values for future time steps of shape (n_steps, n_features).

        Raises
        ------
        ValueError
            If model has not been trained before prediction or if model_type is unsupported.

        Notes
        -----
        Uses recursive prediction strategy with model-specific handling:

        - For LSTM and Transformer models:
            1. Extract the most recent `sequence_length` points as the initial context.
            2. Predict the next time step.
            3. Slide the window: drop the oldest point, append the new prediction.
            4. Repeat for `n_steps`.

        - Critical design choice for LSTM:
            Hidden state is **reset to `None` at every step** to match the training protocol,
            where each batch sample was processed independently with zero-initialized hidden states.
            This ensures consistency between training and inference, improving stability and reproducibility.

        - Output predictions are moved to CPU to avoid GPU memory bloat during long continual loops.
        """
        if not isinstance(historical_data, torch.Tensor):
            historical_data = self.convert_to_tensor(historical_data)

        if model is None:
            raise ValueError("Model must be trained first")

        if len(historical_data) < self.sequence_length:
            raise ValueError(
                f"historical_data has only {len(historical_data)} timesteps, "
                f"but sequence_length = {self.sequence_length} is required."
            )

        # Initialize with historical data (sequence_length, n_features)
        current_sequence = historical_data[-self.sequence_length:].clone().to(self.device)
        predictions = []

        model.eval()

        with torch.no_grad():
            for _ in range(n_steps):
                # Prepare input by adding batch dimension
                input_seq = current_sequence.unsqueeze(0)  # (1, sequence_length, n_features); already on device

                # Generate prediction using trained model with model-specific handling
                if self.model_type == "lstm":
                    # CRITICAL: Use `hidden=None` to match training protocol.
                    # Training used independent sequences with zero-initialized states.
                    # Passing a carried-over hidden would create state-input misalignment.
                    pred, _ = model(input_seq, hidden=None)  # (1, n_features); stateless sliding window
                elif self.model_type == "transformer":
                    # Transformer has no recurrent state — naturally stateless.
                    pred = model(input_seq)  # (1, n_features)
                else:
                    raise ValueError(f"Unsupported model_type: {self.model_type}. Use 'lstm' or 'transformer'.")

                pred = pred.squeeze(0)  # (n_features,), still on device
                predictions.append(pred.cpu())  # only detach & move to CPU for storage, leaves original pred unchanged

                # Update sequence: remove oldest point, append new prediction
                current_sequence = torch.cat([current_sequence[1:], pred.unsqueeze(0)])  # stays on device

        return torch.stack(predictions)  # Stack into (n_steps, n_features)
